goog.provide 'app.references.Store'

goog.require 'goog.events.EventTarget'
goog.require 'app.document.Reference'
goog.require 'app.ReferenceLike'

class app.references.Store extends goog.events.EventTarget

  ###* @enum {string} ###
  @EventType: {
    Change: 'reference-store-change'
  }

  ###* @type {Object} ###
  referenceDocument: null

  ###* @type {Array.<app.document.Reference>} ###
  references: null

  ###* @type {boolean} - state, app must work on latest _rev ###
  canPutNewVersion: true

  ###*
    @constructor
    @param {PouchDB} db
    @param {Function=} randomStringGenerator
    @extends {goog.events.EventTarget}
    @final
  ###
  constructor: (
    @db
    @randomStringGenerator = goog.string.getRandomString
  ) ->
    super()

  ###*
    @param {app.ReferenceLike} object
    @return {app.document.Abbreviation}
  ###
  validateAbbreviation: (object) ->
    object.fullTitle ||= ''
    return `/** @type {app.document.Abbreviation} */ (object)`

  ###*
    @param {app.ReferenceLike} object
    @return {app.document.Citation}
  ###
  validateCitation: (object) ->
    unless object.citation?
      object.citation = ''
    return `/** @type {app.document.Citation} */ (object)`

  ###*
    @param {app.ReferenceLike} object
    @return {app.document.Link}
  ###
  validateLink: (object) ->
    object.url ||= ''
    return `/** @type {app.document.Link} */ (object)`

  ###*
    @param {app.ReferenceLike} object
    @return {app.document.Term}
  ###
  validateTerm: (object) ->
    object.description ||= ''
    return `/** @type {app.document.Term} */ (object)`

  ###*
    Use to get valid reference object from e.g. json.
    @param {Object} object
    @return {app.document.Reference}
  ###
  validateReference: (object) ->
    unless object._id
      object._id = @randomStringGenerator()
    unless object.type
      throw Error 'Reference does not have a type'
    object.title ||= ''
    return `/** @type {app.document.Reference} */ (object)`

  _save: ->
    return unless @canPutNewVersion
    @canPutNewVersion = false
    putPromise = @db.put(@referenceDocument)
    putPromise
      .then(
        @_updateRev
      )
      .then(
        (info) -> console.log info
        (err) -> console.error err
      )

    console.log 'saving references'

  ###*
    Syncs _rev attribute for model with database object.
    @param {{rev: string}} result
  ###
  _updateRev: (result) =>
    console.log 'new rev: ', result
    @referenceDocument._rev = result.rev
    @canPutNewVersion = true

  _notify: ->
    @dispatchEvent app.references.Store.EventType.Change
    console.log 'notifying on changes'

  _saveAndNotify: ->
    @_save()
    @_notify()

  ###*
    @template T
    @param {{type: app.document.Reference.Type}} json
    @param {boolean=} shouldCreate - whether passed object should be valid or
      is just specs of new one
      default to true
    @return {T}
  ###
  makeReference: (json, shouldCreate=true) ->
    object = @validateReference json

    validateFunc = switch object.type
      when app.document.Reference.Type.Abbreviation then @validateAbbreviation
      when app.document.Reference.Type.Citation then @validateCitation
      when app.document.Reference.Type.Link then @validateLink
      when app.document.Reference.Type.Term then @validateTerm
      else console.error('no type in createReference ', object.type)
    object = validateFunc object
    return (object)

  reportError: (err) ->
    console.error(err)

  handlePouchDBError: (err) ->
    if err.status == 404
      @createDocument()
    else
      console.warn 'unhandled pouchdb error: ', err

  ###*
    @return {Array.<app.document.Reference> | Promise}
  ###
  getReferencesOrPromise: ->
    if @referenceDocument
      return @referenceDocument.references
    else
      return @loadReferences()

  ###*
    @param {app.document.Reference.Type} type
  ###
  addReference: (type) ->
    @referenceDocument.references.push {type}
    @_saveAndNotify()

  ###*
    @param {app.document.Reference} ref
    @param {string} newTitle
  ###
  changeTitle: (ref, newTitle) ->
    ref.title = newTitle
    @_saveAndNotify()

  ###*
    @param {app.document.Reference} ref
    @param {Object} changeObject
    @this {app.references.Store} - is passed to event handlers
  ###
  changeReference: (ref, changeObject) =>
    goog.mixin ref, changeObject
    @_saveAndNotify()

  ###*
    Loads references from local/remote database.
    @return {Promise}
  ###
  loadReferences: =>
    getResult = @db.get('docs-references')
    saveResult = getResult.then(
      (document) => @referenceDocument = document
      @handlePouchDBError
    )
    return saveResult

  createDocument: ->
    @referenceDocument = {
      references: @_getFixtures()
      _id: 'docs-references'
    }

  _getFixtures: ->
    return [
        {
          title: 'wikibooks: Paragraph formatting'
          citation: 'LaTeX/Paragraph Formatting - Wikibooks, open books for an open world
En.wikibooks.org, (2014). LaTeX/Paragraph Formatting - Wikibooks, open books for an open world. [online] Available at: http://en.wikibooks.org/wiki/LaTeX/Paragraph_Formatting [Accessed 17 Nov. 2014].'
          type: app.document.Reference.Type.Citation
        }
        {
          title: 'coffeescript.org'
          link: 'http://coffeescript.org/'
          type: app.document.Reference.Type.Link
        }
        {
          title: 'DRY'
          fullTitle: "Don't repeat yourself"
          type: app.document.Reference.Type.Abbreviation
        }
        {
          title: 'functional programming'
          type: app.document.Reference.Type.Term
          description: 'Functional programming is a programming paradigm, a style of building the structure and elements of computer programs, that treats computation as the evaluation of mathematical functions and avoids changing-state and mutable data.'
        }
    ].map (ref) => @makeReference ref, true
